#include "opencv2/opencv.hpp"

#include "stattimer.hpp"
//STimerList st(STimerList::reporterTSV, "stat.txt");
STimerList st;

#ifdef _OPENMP
#include <omp.h>
#define NUM_THREADS_OPENMP num_threads(4)
#endif

#include <emmintrin.h>

//#define LOOP_COLUMN_MAJOR

int b2gray[256];
int g2gray[256];
int r2gray[256];

void highlightFrameDiffMisc(const cv::Mat& input_color,
                            cv::Mat& gray, 
                            cv::Mat& prev_gray,
                            cv::Mat& output_color, int index)
{
    const int width = input_color.cols;
    const int height = input_color.rows;
    
    // convert color
    if (index % 3 == 0) {
        st.start("bgr2gray OpenCV");
        cv::cvtColor(input_color, gray, CV_BGR2GRAY);
        st.stop("bgr2gray OpenCV");
    } else if (index % 3 == 1) {
        st.start("bgr2gray calc");

#ifndef LOOP_COLUMN_MAJOR
        for (int j = 0; j < height; j++) {
            for (int i = 0; i < width; i++) {
#else
        for (int i = 0; i < width; i++) {
            for (int j = 0; j < height; j++) {
#endif
                cv::Vec3b p = input_color.at<cv::Vec3b>(j, i);
                uchar b = p[0];
                uchar g = p[1];
                uchar r = p[2];
                gray.at<uchar>(j, i) = (r * 306 + g * 601 + b * 117) / 1024;
            }
        }
        st.stop("bgr2gray calc");
    } else {
        st.start("bgr2gray lut");

#ifndef LOOP_COLUMN_MAJOR
        for (int j = 0; j < height; j++) {
            for (int i = 0; i < width; i++) {
#else
        for (int i = 0; i < width; i++) {
            for (int j = 0; j < height; j++) {
#endif
                cv::Vec3b p = input_color.at<cv::Vec3b>(j, i);
                uchar b = p[0];
                uchar g = p[1];
                uchar r = p[2];
                gray.at<uchar>(j, i) = (b2gray[b] + g2gray[g] + r2gray[r]) / 1024;
            }
        }
        st.stop("bgr2gray lut");
    }

    // frame diff
    if (index % 3 == 0) {
        st.start("highlight diff OpenCV");
        cv::Mat diff;
        cv::absdiff(gray, prev_gray, diff);
        cv::compare(diff, cv::Scalar(30), diff, cv::CMP_GT);

        std::vector<cv::Mat> channel(3);
        cv::split(input_color, channel);
        channel[0] = gray | diff;
        channel[1] = gray & ~diff;
        channel[2] = gray & ~diff;
        cv::merge(channel, output_color);
        st.stop("highlight diff OpenCV");
    } else if (index % 3 == 1) {
        st.start("highlight diff if-else");
#ifndef LOOP_COLUMN_MAJOR
        for (int j = 0; j < height; j++) {
            for (int i = 0; i < width; i++) {
#else
        for (int i = 0; i < width; i++) {
            for (int j = 0; j < height; j++) {
#endif
                uchar g = gray.at<uchar>(j, i);
                int diff =  g - prev_gray.at<uchar>(j, i);
                if (diff > 30 || diff < -30) {
                    output_color.at<cv::Vec3b>(j, i)[0] = 255;
                    output_color.at<cv::Vec3b>(j, i)[1] = 0;
                    output_color.at<cv::Vec3b>(j, i)[2] = 0;
                } else {
                    output_color.at<cv::Vec3b>(j, i)[0] = g;
                    output_color.at<cv::Vec3b>(j, i)[1] = g;
                    output_color.at<cv::Vec3b>(j, i)[2] = g;
                }
            }
        }
        st.stop("highlight diff if-else");
    } else {
        st.start("highlight diff logical");
#ifndef LOOP_COLUMN_MAJOR
        for (int j = 0; j < height; j++) {
            for (int i = 0; i < width; i++) {
#else
        for (int i = 0; i < width; i++) {
            for (int j = 0; j < height; j++) {
#endif
                uchar g = gray.at<uchar>(j, i);
                int diff =  g - prev_gray.at<uchar>(j, i);
                int active = ((diff > 30 || diff < -30) && 255);
                output_color.at<cv::Vec3b>(j, i)[0] = g | active;
                output_color.at<cv::Vec3b>(j, i)[1] = g & ~active;
                output_color.at<cv::Vec3b>(j, i)[2] = g & ~active;
            }
        }
        st.stop("highlight diff logical");
    }

    // save current frame
    if (index % 3 == 0) {
        st.start("copy OpenCV");
        gray.copyTo(prev_gray);
        st.stop("copy OpenCV");
    } else if (index % 3 == 1) {
        st.start("copy pixelwise");
#ifndef LOOP_COLUMN_MAJOR
        for (int j = 0; j < height; j++) {
            for (int i = 0; i < width; i++) {
#else
        for (int i = 0; i < width; i++) {
            for (int j = 0; j < height; j++) {
#endif
                prev_gray.at<uchar>(j, i) = gray.at<uchar>(j, i);
            }
        }
        st.stop("copy pixelwise");
    } else {
        st.start("copy by memcpy");
        memcpy(prev_gray.data, gray.data, width * height);
        st.stop("copy by memcpy");        
    }
}

void highlightFrameDiffPixelAccess(const cv::Mat& input_color,
                                   cv::Mat& gray, 
                                   cv::Mat& prev_gray,
                                   cv::Mat& output_color, int index)
{
    const int width = input_color.cols;
    const int height = input_color.rows;

    if (1) {
        st.start("bgr2gray ptr");
        for (int j = 0; j < height; j++) {
            const cv::Vec3b *in_ptr = input_color.ptr<cv::Vec3b>(j);
            uchar *g_ptr = gray.ptr<uchar>(j);
            for (int i = 0; i < width; i++) {
                uchar b = in_ptr[i][0];
                uchar g = in_ptr[i][1];
                uchar r = in_ptr[i][2];
                g_ptr[i] = (r * 306 + g * 601 + b * 117) / 1024;
            }
        }
        st.stop("bgr2gray ptr");
    }

    if (index % 2 == 0) {
        st.start("highlight diff ptr");
        for (int j = 0; j < height; j++) {
            uchar *g_ptr = gray.ptr<uchar>(j);
            uchar *prev_g_ptr = prev_gray.ptr<uchar>(j);
            cv::Vec3b *out_ptr = output_color.ptr<cv::Vec3b>(j);
            for (int i = 0; i < width; i++) {
                uchar g = g_ptr[i];
                int diff =  g - prev_g_ptr[i];
                if (diff > 30 || diff < -30) {
                    out_ptr[i][0] = 255;
                    out_ptr[i][1] = 0;
                    out_ptr[i][2] = 0;
                } else {
                    out_ptr[i][0] = g;
                    out_ptr[i][1] = g;
                    out_ptr[i][2] = g;
                }
            }
        }
        st.stop("highlight diff ptr");
    } else {
        st.start("highlight diff ptr bulk");
        for (int j = 0; j < height; j++) {
            uchar *g_ptr = gray.ptr<uchar>(j);
            uchar *prev_g_ptr = prev_gray.ptr<uchar>(j);
            cv::Vec3b *out_ptr = output_color.ptr<cv::Vec3b>(j);
            for (int i = 0; i < width; i++) {
                uchar g = g_ptr[i];
                int diff =  g - prev_g_ptr[i];
                if (diff > 30 || diff < -30) {
                    out_ptr[i] = cv::Vec3b(255, 0, 0);
                } else {
                    out_ptr[i] = cv::Vec3b(g, g, g);
                }
            }
        }
        st.stop("highlight diff ptr bulk");
    }

    memcpy(prev_gray.data, gray.data, width * height);
}

void highlightFrameDiffLoopFusion(const cv::Mat& input_color,
                                  cv::Mat& gray, 
                                  cv::Mat& prev_gray,
                                  cv::Mat& output_color, int index)
{
    const int width = input_color.cols;
    const int height = input_color.rows;

    if (index % 3 == 0) {
        st.start("loop not fused");
        for (int j = 0; j < height; j++) {
            const cv::Vec3b *in_ptr = input_color.ptr<cv::Vec3b>(j);
            uchar *g_ptr = gray.ptr<uchar>(j);
            for (int i = 0; i < width; i++) {
                uchar b = in_ptr[i][0];
                uchar g = in_ptr[i][1];
                uchar r = in_ptr[i][2];
                g_ptr[i] = (r * 306 + g * 601 + b * 117) / 1024;
            }
        }
        for (int j = 0; j < height; j++) {
            uchar *g_ptr = gray.ptr<uchar>(j);
            uchar *prev_g_ptr = prev_gray.ptr<uchar>(j);
            cv::Vec3b *out_ptr = output_color.ptr<cv::Vec3b>(j);
            for (int i = 0; i < width; i++) {
                uchar grayval = g_ptr[i];
                int diff =  grayval - prev_g_ptr[i];
                if (diff > 30 || diff < -30) {
                    out_ptr[i][0] = 255;
                    out_ptr[i][1] = 0;
                    out_ptr[i][2] = 0;
                } else {
                    out_ptr[i][0] = grayval;
                    out_ptr[i][1] = grayval;
                    out_ptr[i][2] = grayval;
                }
            }
        }
        memcpy(prev_gray.data, gray.data, width * height);
        st.stop("loop not fused");
    } else if (index % 3 == 1) {
        st.start("loop fused");
        for (int j = 0; j < height; j++) {
            const cv::Vec3b *in_ptr = input_color.ptr<cv::Vec3b>(j);
            uchar *g_ptr = gray.ptr<uchar>(j);
            uchar *prev_g_ptr = prev_gray.ptr<uchar>(j);
            cv::Vec3b *out_ptr = output_color.ptr<cv::Vec3b>(j);
            for (int i = 0; i < width; i++) {
                uchar b = in_ptr[i][0];
                uchar g = in_ptr[i][1];
                uchar r = in_ptr[i][2];
                g_ptr[i] = (r * 306 + g * 601 + b * 117) / 1024;

                uchar grayval = g_ptr[i];
                int diff =  grayval - prev_g_ptr[i];
                if (diff > 30 || diff < -30) {
                    out_ptr[i][0] = 255;
                    out_ptr[i][1] = 0;
                    out_ptr[i][2] = 0;
                } else {
                    out_ptr[i][0] = grayval;
                    out_ptr[i][1] = grayval;
                    out_ptr[i][2] = grayval;
                }

                prev_g_ptr[i] = grayval;
            }
        }
        st.stop("loop fused");
    } else {
        st.start("loop fused except memcpy");
        for (int j = 0; j < height; j++) {
            const cv::Vec3b *in_ptr = input_color.ptr<cv::Vec3b>(j);
            uchar *g_ptr = gray.ptr<uchar>(j);
            uchar *prev_g_ptr = prev_gray.ptr<uchar>(j);
            cv::Vec3b *out_ptr = output_color.ptr<cv::Vec3b>(j);
            for (int i = 0; i < width; i++) {
                uchar b = in_ptr[i][0];
                uchar g = in_ptr[i][1];
                uchar r = in_ptr[i][2];
                g_ptr[i] = (r * 306 + g * 601 + b * 117) / 1024;

                uchar grayval = g_ptr[i];
                int diff =  grayval - prev_g_ptr[i];
                if (diff > 30 || diff < -30) {
                    out_ptr[i][0] = 255;
                    out_ptr[i][1] = 0;
                    out_ptr[i][2] = 0;
                } else {
                    out_ptr[i][0] = grayval;
                    out_ptr[i][1] = grayval;
                    out_ptr[i][2] = grayval;
                }
            }
        }
        memcpy(prev_gray.data, gray.data, width * height);        
        st.stop("loop fused except memcpy");
    }

}


void highlightFrameDiffLoopUnroll(const cv::Mat& input_color,
                                  cv::Mat& gray, 
                                  cv::Mat& prev_gray,
                                  cv::Mat& output_color, int index)
{
    st.start("loop unrolling");

    const int width = input_color.cols;
    const int height = input_color.rows;

    if (width % 4 != 0) {
        fprintf(stderr, "width must be multiples of 4: width = %d\n", width);
        return;
    }

    {
        for (int j = 0; j < height; j++) {
            const cv::Vec3b *in_ptr = input_color.ptr<cv::Vec3b>(j);
            uchar *g_ptr = gray.ptr<uchar>(j);
            uchar *prev_g_ptr = prev_gray.ptr<uchar>(j);
            cv::Vec3b *out_ptr = output_color.ptr<cv::Vec3b>(j);

            for (int i = 0; i < width; i += 4) {
                // 0th pixel
                uchar b = in_ptr[i][0];
                uchar g = in_ptr[i][1];
                uchar r = in_ptr[i][2];
                g_ptr[i] = (r * 306 + g * 601 + b * 117) / 1024;

                uchar grayval = g_ptr[i];
                int diff =  grayval - prev_g_ptr[i];
                if (diff > 30 || diff < -30) {
                    out_ptr[i][0] = 255;
                    out_ptr[i][1] = 0;
                    out_ptr[i][2] = 0;
                } else {
                    out_ptr[i][0] = grayval;
                    out_ptr[i][1] = grayval;
                    out_ptr[i][2] = grayval;
                }

                // 1st pixel
                b = in_ptr[i + 1][0];
                g = in_ptr[i + 1][1];
                r = in_ptr[i + 1][2];
                g_ptr[i + 1] = (r * 306 + g * 601 + b * 117) / 1024;

                grayval = g_ptr[i + 1];
                diff =  grayval - prev_g_ptr[i + 1];
                if (diff > 30 || diff < -30) {
                    out_ptr[i + 1][0] = 255;
                    out_ptr[i + 1][1] = 0;
                    out_ptr[i + 1][2] = 0;
                } else {
                    out_ptr[i + 1][0] = grayval;
                    out_ptr[i + 1][1] = grayval;
                    out_ptr[i + 1][2] = grayval;
                }

                // 2nd pixel
                b = in_ptr[i + 2][0];
                g = in_ptr[i + 2][1];
                r = in_ptr[i + 2][2];
                g_ptr[i + 2] = (r * 306 + g * 601 + b * 117) / 1024;

                grayval = g_ptr[i + 2];
                diff =  grayval - prev_g_ptr[i + 2];
                if (diff > 30 || diff < -30) {
                    out_ptr[i + 2][0] = 255;
                    out_ptr[i + 2][1] = 0;
                    out_ptr[i + 2][2] = 0;
                } else {
                    out_ptr[i + 2][0] = grayval;
                    out_ptr[i + 2][1] = grayval;
                    out_ptr[i + 2][2] = grayval;
                }

                // 3rd pixel
                b = in_ptr[i + 3][0];
                g = in_ptr[i + 3][1];
                r = in_ptr[i + 3][2];
                g_ptr[i + 3] = (r * 306 + g * 601 + b * 117) / 1024;

                grayval = g_ptr[i + 3];
                diff =  grayval - prev_g_ptr[i + 3];
                if (diff > 30 || diff < -30) {
                    out_ptr[i + 3][0] = 255;
                    out_ptr[i + 3][1] = 0;
                    out_ptr[i + 3][2] = 0;
                } else {
                    out_ptr[i + 3][0] = grayval;
                    out_ptr[i + 3][1] = grayval;
                    out_ptr[i + 3][2] = grayval;
                }
            }

        }
        memcpy(prev_gray.data, gray.data, width * height);        
    }
    st.stop("loop unrolling");
}


void highlightFrameDiffOpenMP(const cv::Mat& input_color,
                              cv::Mat& gray, 
                              cv::Mat& prev_gray,
                              cv::Mat& output_color, int index)
{
    st.start("OpenMP");

    const int width = input_color.cols;
    const int height = input_color.rows;

    if (width % 4 != 0) {
        fprintf(stderr, "width must be multiples of 4: width = %d\n", width);
        return;
    }

    {
        st.start("bgr2gray OpenMP");
#ifdef _OPENMP
#pragma omp parallel for NUM_THREADS_OPENMP
#endif
        for (int j = 0; j < height; j++) {
            const cv::Vec3b *in_ptr = input_color.ptr<cv::Vec3b>(j);
            uchar *g_ptr = gray.ptr<uchar>(j);
            for (int i = 0; i < width; i += 4) {
                // 0th pixel
                uchar b = in_ptr[i][0];
                uchar g = in_ptr[i][1];
                uchar r = in_ptr[i][2];
                g_ptr[i] = (r * 306 + g * 601 + b * 117) / 1024;

                // 1st pixel
                b = in_ptr[i + 1][0];
                g = in_ptr[i + 1][1];
                r = in_ptr[i + 1][2];
                g_ptr[i + 1] = (r * 306 + g * 601 + b * 117) / 1024;

                // 2nd pixel
                b = in_ptr[i + 2][0];
                g = in_ptr[i + 2][1];
                r = in_ptr[i + 2][2];
                g_ptr[i + 2] = (r * 306 + g * 601 + b * 117) / 1024;

                // 3rd pixel
                b = in_ptr[i + 3][0];
                g = in_ptr[i + 3][1];
                r = in_ptr[i + 3][2];
                g_ptr[i + 3] = (r * 306 + g * 601 + b * 117) / 1024;
            }
        }
        st.stop("bgr2gray OpenMP");

        st.start("highlight diff OpenMP");
#ifdef _OPENMP
#pragma omp parallel for NUM_THREADS_OPENMP
#endif
        for (int j = 0; j < height; j++) {
            uchar *g_ptr = gray.ptr<uchar>(j);
            uchar *prev_g_ptr = prev_gray.ptr<uchar>(j);
            cv::Vec3b *out_ptr = output_color.ptr<cv::Vec3b>(j);

            for (int i = 0; i < width; i += 4) {
                // 0th pixel
                uchar grayval = g_ptr[i];
                int diff =  grayval - prev_g_ptr[i];
                if (diff > 30 || diff < -30) {
                    out_ptr[i][0] = 255;
                    out_ptr[i][1] = 0;
                    out_ptr[i][2] = 0;
                } else {
                    out_ptr[i][0] = grayval;
                    out_ptr[i][1] = grayval;
                    out_ptr[i][2] = grayval;
                }

                // 1st pixel
                grayval = g_ptr[i + 1];
                diff =  grayval - prev_g_ptr[i + 1];
                if (diff > 30 || diff < -30) {
                    out_ptr[i + 1][0] = 255;
                    out_ptr[i + 1][1] = 0;
                    out_ptr[i + 1][2] = 0;
                } else {
                    out_ptr[i + 1][0] = grayval;
                    out_ptr[i + 1][1] = grayval;
                    out_ptr[i + 1][2] = grayval;
                }

                // 2nd pixel
                grayval = g_ptr[i + 2];
                diff =  grayval - prev_g_ptr[i + 2];
                if (diff > 30 || diff < -30) {
                    out_ptr[i + 2][0] = 255;
                    out_ptr[i + 2][1] = 0;
                    out_ptr[i + 2][2] = 0;
                } else {
                    out_ptr[i + 2][0] = grayval;
                    out_ptr[i + 2][1] = grayval;
                    out_ptr[i + 2][2] = grayval;
                }

                // 3rd pixel
                grayval = g_ptr[i + 3];
                diff =  grayval - prev_g_ptr[i + 3];
                if (diff > 30 || diff < -30) {
                    out_ptr[i + 3][0] = 255;
                    out_ptr[i + 3][1] = 0;
                    out_ptr[i + 3][2] = 0;
                } else {
                    out_ptr[i + 3][0] = grayval;
                    out_ptr[i + 3][1] = grayval;
                    out_ptr[i + 3][2] = grayval;
                }
            }
        }
        st.stop("highlight diff OpenMP");

        memcpy(prev_gray.data, gray.data, width * height);        
    }
    st.stop("OpenMP");
}

void highlightFrameDiffSSE(const cv::Mat& input_color,
                           cv::Mat& gray, 
                           cv::Mat& prev_gray,
                           cv::Mat& output_color, int index)
{
    st.start("OpenMP & SSE");

    const int width = input_color.cols;
    const int height = input_color.rows;

    if (width % 16 != 0) {
        fprintf(stderr, "width must be multiples of 16: width = %d\n", width);
        return;
    }

    {
        st.start("bgr2gray OpenMP & SSE");

#ifdef _OPENMP
#pragma omp parallel for NUM_THREADS_OPENMP
#endif
        for (int j = 0; j < height - 1; j++) {
            const cv::Vec3b *in_ptr = input_color.ptr<cv::Vec3b>(j);
            uchar *g_ptr = gray.ptr<uchar>(j);
            uchar tmpbuf[16];
            for (int i = 0; i < width; i += 4) {

                __m128i data = _mm_loadu_si128((__m128i *)(in_ptr + i));
                // BGRBGRBGRBGR(BGRB)

                __m128i bmask = _mm_setr_epi8(0,-1,-1,-1,
                                              3,-1,-1,-1,
                                              6,-1,-1,-1,
                                              9,-1,-1,-1);
                __m128i b4 = _mm_shuffle_epi8(data, bmask);
                __m128i bcoeff = _mm_setr_epi32(306, 306, 306, 306);
                b4 = _mm_mullo_epi32(b4, bcoeff);

                __m128i gmask = _mm_setr_epi8( 1,-1,-1,-1,
                                               4,-1,-1,-1,
                                               7,-1,-1,-1,
                                              10,-1,-1,-1);
                __m128i g4 = _mm_shuffle_epi8(data, gmask);
                __m128i gcoeff = _mm_setr_epi32(601, 601, 601, 601);
                g4 = _mm_mullo_epi32(g4, gcoeff);

                __m128i rmask = _mm_setr_epi8( 2,-1,-1,-1,
                                               5,-1,-1,-1,
                                               8,-1,-1,-1,
                                              11,-1,-1,-1);
                __m128i r4 = _mm_shuffle_epi8(data, rmask);
                __m128i rcoeff = _mm_setr_epi32(117, 117, 117, 117);
                r4 = _mm_mullo_epi32(r4, rcoeff);

                __m128i grayval = _mm_add_epi32(b4, g4);
                grayval = _mm_add_epi32(grayval, r4);
                grayval = _mm_srli_epi32(grayval, 10);

                _mm_store_si128((__m128i *)tmpbuf, grayval);
                g_ptr[i + 0] = tmpbuf[0];
                g_ptr[i + 1] = tmpbuf[4];
                g_ptr[i + 2] = tmpbuf[8];
                g_ptr[i + 3] = tmpbuf[12];
            }
        }
        { /* To avoid buffer overrun, the last line is not processed by SSE
           * (because we read 16 bytes while we process 4 pixels (= 12 bytes))
           */
            const cv::Vec3b *in_ptr = input_color.ptr<cv::Vec3b>(height - 1);
            uchar *g_ptr = gray.ptr<uchar>(height - 1);
            for (int i = 0; i < width; i++) {
                uchar b = in_ptr[i][0];
                uchar g = in_ptr[i][1];
                uchar r = in_ptr[i][2];
                g_ptr[i] = (r * 306 + g * 601 + b * 117) / 1024;
            }
        }
        st.stop("bgr2gray OpenMP & SSE");

        st.start("highlight diff OpenMP & SSE");
#ifdef _OPENMP
#pragma omp parallel for NUM_THREADS_OPENMP
#endif
        for (int j = 0; j < height; j++) {
            uchar *g_ptr = gray.ptr<uchar>(j);
            uchar *prev_g_ptr = prev_gray.ptr<uchar>(j);
            cv::Vec3b *out_ptr = output_color.ptr<cv::Vec3b>(j);

            uchar buf_bchan[16];
            uchar buf_grchan[16];
            for (int i = 0; i < width; i += 16) {
                __m128i g = _mm_loadu_si128((__m128i *)(g_ptr + i));
                __m128i prev_g = _mm_loadu_si128((__m128i *)(prev_g_ptr + i));
                __m128i zero = _mm_setzero_si128();
                __m128i thresh = _mm_set_epi16(30, 30, 30, 30, 30, 30, 30, 30);

                //// lo 8 bytes
                // g - prev_g
                __m128i gs = _mm_unpacklo_epi8(g, zero);
                __m128i prev_gs = _mm_unpacklo_epi8(prev_g, zero);
                __m128i diff = _mm_sub_epi16(gs, prev_gs);

                // abs()
                __m128i bitmask = _mm_srai_epi16(diff, 15);
                diff = _mm_xor_si128(diff, bitmask);
                diff = _mm_sub_epi16(diff, bitmask);

                // > 30 ?
                diff = _mm_cmpgt_epi16(diff, thresh);
                __m128i flag = _mm_packs_epi16(diff, zero);

                //// hi 8 bytes
                // g - prev_g
                gs = _mm_unpackhi_epi8(g, zero);
                prev_gs = _mm_unpackhi_epi8(prev_g, zero);
                diff = _mm_sub_epi16(gs, prev_gs);

                // abs()
                bitmask = _mm_srai_epi16(diff, 15);
                diff = _mm_xor_si128(diff, bitmask);
                diff = _mm_sub_epi16(diff, bitmask);

                // > 30 ?
                diff = _mm_cmpgt_epi16(diff, thresh);
                diff = _mm_packs_epi16(diff, zero);
                diff = _mm_slli_si128(diff, 8);
                flag = _mm_or_si128(flag, diff);

                __m128i bchan = _mm_or_si128(flag, g);
                _mm_store_si128((__m128i *)buf_bchan, bchan);
                __m128i grchan = _mm_andnot_si128(flag, g);
                _mm_store_si128((__m128i *)buf_grchan, grchan);

                for (int k = 0; k < 16; k++) {
                    out_ptr[i + k][0] = buf_bchan[k];
                    out_ptr[i + k][1] = buf_grchan[k];
                    out_ptr[i + k][2] = buf_grchan[k];

                    // out_ptr[i + k + 1][0] = buf_bchan[k + 1];
                    // out_ptr[i + k + 1][1] = buf_grchan[k + 1];
                    // out_ptr[i + k + 1][2] = buf_grchan[k + 1];

                    // out_ptr[i + k + 2][0] = buf_bchan[k + 2];
                    // out_ptr[i + k + 2][1] = buf_grchan[k + 2];
                    // out_ptr[i + k + 2][2] = buf_grchan[k + 2];

                    // out_ptr[i + k + 3][0] = buf_bchan[k + 3];
                    // out_ptr[i + k + 3][1] = buf_grchan[k + 3];
                    // out_ptr[i + k + 3][2] = buf_grchan[k + 3];
                }
            }
        }
        st.stop("highlight diff OpenMP & SSE");

        memcpy(prev_gray.data, gray.data, width * height);        
    }
    st.stop("OpenMP & SSE");
}

int main()
{
    // setup lookup table
    for (int i = 0; i < 256; i++) {
        b2gray[i] = 117 * i;
        g2gray[i] = 601 * i;
        r2gray[i] = 306 * i;
    }

    cv::namedWindow("display", cv::WINDOW_AUTOSIZE);
    cv::VideoCapture cap(0);
    cv::Mat input;
    cap >> input;
    cv::Mat gray(input.rows, input.cols, CV_8U);
    cv::Mat prev_gray(input.rows, input.cols, CV_8U);
    cv::Mat output(input.rows, input.cols, CV_8UC3);

    int index = 0;
    while (1) {
        cap >> input;

//        highlightFrameDiffMisc(input, gray, prev_gray, output, index);
        highlightFrameDiffPixelAccess(input, gray, prev_gray, output, index);
//        highlightFrameDiffLoopFusion(input, gray, prev_gray, output, index);
//        highlightFrameDiffLoopUnroll(input, gray, prev_gray, output, index);
//        highlightFrameDiffOpenMP(input, gray, prev_gray, output, index);
//        highlightFrameDiffSSE(input, gray, prev_gray, output, index);

        cv::imshow("display", output);
        int key;
        if ((key = cv::waitKey(30)) > 0) {
            if (key == 'q' || key == 0x1b) {
                break;
            }
        }

        index++;
    }

}
